You are being asked a question that requires a numerical response 
in the form of an integer or decimal (e.g., -12, 0, 1, 2, 3.45, ...).
        
Your response must be in the following format:

{% if include_comment %}
    {"answer": "<your numerical answer here>", "comment": "<your explanation here>"}
{% else %}
    {"answer": "<your numerical answer here>"}
{% endif %}

You must only include an integer or decimal in the quoted "answer" part of your response.
        
Here is an example of a valid response:
{% if include_comment %}
    {"answer": "100", "comment": "This is my explanation..."}
{% else %}
    {"answer": "100"}
{% endif %}
        
Here is an example of a response that is invalid because the "answer" includes words:
{"answer": "I don't know.", ...}

If your response is equivalent to zero, your formatted response should look like this:
{% if include_comment %}
    {"answer": "0", "comment": "This is my explanation..."}
{% else %}
    {"answer": "0"}
{% endif %}

You are being asked the following question: {{question_text}}
{% if min_value is not none %}
    Minimum answer value: {{min_value}}
{% endif %}
{% if max_value is not none %}
    Maximum answer value: {{max_value}}
{% endif %}