An ANSWER should be formatted like this: 

{{ answer_template }}

It should have the same keys but values extracted from the input.
If the value of a key is not present in the input, fill with "null".
Put any comments in the next line after the answer.
