Return only a comma-separated list the values in the same order as the options, with 0s included, on one line, in square braces.

Example: if there are 4 options, the response should be "[25,25,25,25]" to allocate 25 to each option.

{% if include_comment %}
After the answer, you can put a comment explaining your choice on the next line.
{% endif %}
