from .pdf import PdfMethods
from .docx import DocxMethods
from .png import PngMethods
from .txt import TxtMethods
from .html import HtmlMethods
from .md import MarkdownMethods
from .csv import CsvMethods
from .json import JsonMethods
from .sql import SqlMethods
from .pptx import PptxMethods
from .latex import LaTeXMethods
from .py import PyMethods
from .sqlite import SQLiteMethods
from .jpeg import JpegMethods
