<style>
    td {
        padding: 0 10px; /* This applies the padding uniformly to all td elements */
    }
</style>

<h3>Exceptions Report</h3>
<p>
    This report summarizes exceptions encountered in the job that was run.
</p>

<table border="1">
    <tbody>
        <tr>
            <td>Total interviews</td>
            <td>{{ interviews|length }}</td>
        </tr>
        <tr>
            <td>Interviews with exceptions</td>
            <td>{{ num_exceptions }}</td>
        </tr>
    </tbody>
</table>
<p>
    An interview is the result of one survey, taken by one agent, with one model and one scenario (if any).
</p>
